from ..helpers import B2LuigiTestCase

import b2luigi


class RequiresTestCase(B2LuigiTestCase):
    def test_requires(self):
        class TaskA(b2luigi.Task):
            some_parameter = b2luigi.IntParameter()
            some_other_parameter = b2luigi.IntParameter()

            def output(self):
                yield self.add_to_output("test.txt")

        @b2luigi.requires(TaskA, some_parameter=3)
        class TaskB(b2luigi.Task):
            another_parameter = b2luigi.IntParameter()

            def output(self):
                yield self.add_to_output("out.dat")

        task = TaskB(some_other_parameter=1, another_parameter=42)
        self.assertEqual(sorted(task.get_param_names()), ["another_parameter", "some_other_parameter"])
        self.assertEqual(task.another_parameter, 42)
        self.assertEqual(task.some_other_parameter, 1)

        self.assertTrue(
            task.get_output_file_name("out.dat").endswith("results/some_other_parameter=1/another_parameter=42/out.dat")
        )

        input_files = task.get_input_file_names("test.txt")
        self.assertEqual(len(input_files), 1)
        self.assertTrue(input_files[0].endswith("results/some_parameter=3/some_other_parameter=1/test.txt"))

        required_task = next(task.requires())
        self.assertEqual(sorted(required_task.get_param_names()), ["some_other_parameter", "some_parameter"])
        self.assertEqual(required_task.some_parameter, 3)
        self.assertEqual(required_task.some_other_parameter, 1)
        self.assertTrue(
            required_task.get_output_file_name("test.txt").endswith(
                "results/some_parameter=3/some_other_parameter=1/test.txt"
            )
        )

    def test_requires_multiple_tasks(self):
        class TaskA(b2luigi.Task):
            some_parameter = b2luigi.IntParameter()

            def output(self):
                yield self.add_to_output("test.txt")

        class TaskB(b2luigi.Task):
            some_other_parameter = b2luigi.IntParameter()

            def output(self):
                yield self.add_to_output("test.txt")

        @b2luigi.requires(TaskA, TaskB)
        class TaskC(b2luigi.Task):
            another_parameter = b2luigi.IntParameter()

            def output(self):
                yield self.add_to_output("out.dat")

        task = TaskC(some_parameter=1, some_other_parameter=2, another_parameter=42)
        self.assertEqual(
            sorted(task.get_param_names()), ["another_parameter", "some_other_parameter", "some_parameter"]
        )
        self.assertEqual(task.some_parameter, 1)
        self.assertEqual(task.some_other_parameter, 2)
        self.assertEqual(task.another_parameter, 42)

        self.assertTrue(
            task.get_output_file_name("out.dat").endswith(
                "results/some_parameter=1/some_other_parameter=2/another_parameter=42/out.dat"
            )
        )


class InheritsTestCase(B2LuigiTestCase):
    def test_inherits(self):
        class TaskA(b2luigi.Task):
            some_parameter = b2luigi.IntParameter()
            some_other_parameter = b2luigi.IntParameter()

            def output(self):
                yield self.add_to_output("test.txt")

        @b2luigi.inherits(TaskA, without="some_other_parameter")
        class TaskB(b2luigi.Task):
            another_parameter = b2luigi.IntParameter()

            def requires(self):
                for my_other_parameter in range(10):
                    yield self.clone(TaskA, some_other_parameter=my_other_parameter)

            def run(self):
                # somehow merge the output of TaskA to create "out.dat"
                pass

            def output(self):
                yield self.add_to_output("out.dat")

        task = TaskB(some_parameter=23, another_parameter=42)
        self.assertEqual(task.get_param_names(), ["some_parameter", "another_parameter"])
        self.assertEqual(task.another_parameter, 42)
        self.assertEqual(task.some_parameter, 23)
        self.assertFalse(hasattr(task, "some_other_parameter"))

        self.assertTrue(
            task.get_output_file_name("out.dat").endswith("results/some_parameter=23/another_parameter=42/out.dat")
        )

        input_files = task.get_input_file_names("test.txt")
        self.assertEqual(len(input_files), 10)
        for my_other_parameter in range(10):
            self.assertTrue(
                input_files[my_other_parameter].endswith(
                    f"results/some_parameter=23/some_other_parameter={my_other_parameter}/test.txt"
                )
            )

        required_tasks = list(task.requires())
        for some_other_parameter_values, required_task in enumerate(required_tasks):
            self.assertEqual(sorted(required_task.get_param_names()), ["some_other_parameter", "some_parameter"])
            self.assertEqual(required_task.some_other_parameter, some_other_parameter_values)
            self.assertEqual(required_task.some_parameter, 23)
            self.assertTrue(
                required_task.get_output_file_name("test.txt").endswith(
                    f"results/some_parameter=23/some_other_parameter={some_other_parameter_values}/test.txt"
                )
            )

    def test_inherits_multiple_tasks(self):
        class TaskA(b2luigi.Task):
            some_parameter = b2luigi.IntParameter()

            def output(self):
                yield self.add_to_output("testA.txt")

        class TaskB(b2luigi.Task):
            some_other_parameter = b2luigi.IntParameter()

            def output(self):
                yield self.add_to_output("testB.txt")

        @b2luigi.inherits(TaskA, TaskB, without="some_other_parameter")
        class TaskC(b2luigi.Task):
            another_parameter = b2luigi.IntParameter()

            def requires(self):
                yield self.clone(TaskA, some_parameter=self.some_parameter)
                for my_other_parameter in range(10):
                    yield self.clone(TaskB, some_other_parameter=my_other_parameter)

            def run(self):
                # somehow merge the output of TaskA to create "out.dat"
                pass

            def output(self):
                yield self.add_to_output("out.dat")

        task = TaskC(some_parameter=23, another_parameter=42)
        self.assertEqual(task.get_param_names(), ["some_parameter", "another_parameter"])
        self.assertEqual(task.another_parameter, 42)
        self.assertEqual(task.some_parameter, 23)
        self.assertFalse(hasattr(task, "some_other_parameter"))

        self.assertTrue(
            task.get_output_file_name("out.dat").endswith("results/some_parameter=23/another_parameter=42/out.dat")
        )

        input_files = task.get_input_file_names("testB.txt")
        self.assertEqual(len(input_files), 10)
        for my_other_parameter in range(10):
            self.assertTrue(
                input_files[my_other_parameter].endswith(f"results/some_other_parameter={my_other_parameter}/testB.txt")
            )



"""
Test helper functions for :py:class:`SlurmProcess`.
"""

import pathlib

import subprocess
import unittest
from unittest import mock

import b2luigi
from b2luigi.batch.processes.slurm import SlurmJobStatusCache, SlurmProcess

from ..helpers import B2LuigiTestCase
from .batch_task_1 import MyTask


class TestSlurmCreateSubmitFile(B2LuigiTestCase):
    def _get_slurm_submit_file_string(self, task):
        # the _create_slurm_submit_file is a method of the ``SlurmProcess`` class, but it only uses its
        # class to obtain ``self.task``, so it's sufficient to provide a mock class for ``self``, `
        slurm_mock_process = mock.Mock()
        task.get_task_file_dir = lambda: self.test_dir
        task.get_log_file_dir = lambda: self.test_dir
        slurm_mock_process.task = task
        #  create submit file
        SlurmProcess._create_slurm_submit_file(slurm_mock_process)
        # read submit file and return string
        submit_file_path = pathlib.Path(self.test_dir) / "slurm_parameters.sh"
        with open(submit_file_path, "r") as submit_file:
            return submit_file.read()

    def test_minimal_submit_file(self):
        """
        Minimal submit file should have expected shape:
            #!/usr/bin/bash
            #SBATCH --output=...
            #SBATCH --error=...
            exec executable_wrapper.sh
        """
        task = MyTask("some_parameter_a")
        submit_file_lines = self._get_slurm_submit_file_string(task).splitlines()
        self.assertIn("#!/usr/bin/bash", submit_file_lines[0])
        self.assertIn("#SBATCH --output=", submit_file_lines[1])
        self.assertIn("#SBATCH --error=", submit_file_lines[2])
        self.assertEqual(
            f"exec {(pathlib.Path(self.test_dir)/'executable_wrapper.sh').resolve()}", submit_file_lines[3]
        )

    def test_not_setting_job_name(self):
        task = MyTask("some_parameter_b")
        submit_file_string = self._get_slurm_submit_file_string(task)
        self.assertNotIn("job-name", submit_file_string)

    def test_set_job_name_via_task_attribute(self):
        task = MyTask("some_parameter_c")
        task.job_name = "some_job_name"
        submit_file_lines = self._get_slurm_submit_file_string(task).splitlines()
        self.assertIn("#SBATCH --job-name=some_job_name", submit_file_lines)

        b2luigi.set_setting("job_name", "some_job_name")
        submit_file_lines = self._get_slurm_submit_file_string(task).splitlines()
        b2luigi.clear_setting("job_name")
        self.assertIn("#SBATCH --job-name=some_job_name", submit_file_lines)

    def test_set_job_name_is_overriden_by_slurm_settings(self):
        """
        ``job_name`` is a global setting, but if the ``job-name`` is set explicitly via the settings, we
        want that to override the global setting
        """
        task = MyTask("some_parameter_d")
        task.job_name = "job_name_global"
        slurm_settings = {"job-name": "job_name_slurm"}
        b2luigi.set_setting("slurm_settings", slurm_settings)
        submit_file_lines = self._get_slurm_submit_file_string(task).splitlines()
        b2luigi.clear_setting("slurm_settings")
        self.assertNotIn("#SBATCH --job-name=job_name_global", submit_file_lines)
        self.assertIn("#SBATCH --job-name=job_name_slurm", submit_file_lines)


class TestSlurmJobStatusCache(unittest.TestCase):
    def setUp(self):
        # Slurm output formatted when using squeue and sacct
        mock_squeue_sacct_status_string = "12344 COMPLETED\n12356 RUNNING\n13253 FAILED\n"
        self.mock_status_string = mock_squeue_sacct_status_string.encode()
        # Incorrect sqeue status
        mock_incorrect_squeue_sacct_status_string = "IncorrectString"
        self.mock_incorrect_sqeue_sacct_status_string = mock_incorrect_squeue_sacct_status_string.encode()
        # Slurm output formatted when using scontrol
        mock_scontrol_status_string = """
            JobId=142889 JobName=slurm_parameters.sh
            UserId=test JobState=RUNNING
        """
        self.mock_scontrol_status_string = mock_scontrol_status_string.encode()
        # Empty encoded string for when squeue returns nothing
        self.mock_empty_squeue_status_string = "".encode()
        self.slurm_job_status_cache = SlurmJobStatusCache()

    @mock.patch("subprocess.check_output")
    def test_ask_for_job_status_does_2_retries(self, mock_check_output):
        """Test the ``_ask_for_job_status`` recovers after two squeue failures."""

        # make check_output fail 2 times before  return status dict
        n_fail = 2
        mock_check_output.side_effect = n_fail * [subprocess.CalledProcessError(1, ["mock", "command"])] + [
            self.mock_status_string
        ]
        self.slurm_job_status_cache._ask_for_job_status()
        self.assertEqual(mock_check_output.call_count, 3)

    @mock.patch("subprocess.check_output")
    def test_ask_for_job_status_no_retries_on_success(self, mock_check_output):
        """Test the ``_ask_for_job_status`` is only called once (no retries) when everything works."""

        # make check_output fail 2 times before  return status dict
        mock_check_output.return_value = self.mock_status_string
        self.slurm_job_status_cache._ask_for_job_status()
        self.assertEqual(mock_check_output.call_count, 1)

    @mock.patch("subprocess.check_output")
    def test_ask_for_job_status_fails_after_4_condor_q_failures(self, mock_check_output):
        """Test the ``_ask_for_job_status`` does not do more than 3 retries"""

        # make check_output fail 2 times before  return status dict
        n_fail = 4
        mock_check_output.side_effect = n_fail * [subprocess.CalledProcessError(1, ["mock", "command"])] + [
            self.mock_status_string
        ]
        with self.assertRaises(subprocess.CalledProcessError):
            self.slurm_job_status_cache._ask_for_job_status()

    @mock.patch("subprocess.check_output")
    def test_ask_for_job_with_scontrol_no_retries_on_success(self, mock_check_output):
        """Test that when a system has no sacct, the scontrol method works on first try"""

        # make _check_sacct_is_active_on_server return False to enable scontrol command
        self.slurm_job_status_cache._check_if_sacct_is_disabled_on_server = lambda: True
        self.assertEqual(self.slurm_job_status_cache._check_if_sacct_is_disabled_on_server(), True)
        # Make the first subprocess.check_output call return an empty string to signify
        # there is no output from squeue. Since the _check_sacct_is_active_on_server returns
        # False the second call of subprocess.check_output will be using scontrol
        mock_check_output.side_effect = [self.mock_empty_squeue_status_string] + [self.mock_scontrol_status_string]
        # Run assertion test
        self.slurm_job_status_cache._ask_for_job_status()
        self.assertEqual(mock_check_output.call_count, 1)

    @mock.patch("subprocess.check_output")
    def test_ask_for_job_status_with_scontrol_does_2_retries(self, mock_check_output):
        """Test that when a system has no sacct, scontrol method works on third try"""
        # make check_output fail 2 times before  return status dict
        n_fail = 2
        # After 2 retries, make the first successful subprocess.check_output call return an
        # empty string to signify there is no output from squeue. Since the
        # _check_sacct_is_active_on_server returns False the second call of subprocess.check_output
        # will be using scontrol
        mock_check_output.side_effect = (
            n_fail * [subprocess.CalledProcessError(1, ["mock", "command"])]
            + [self.mock_empty_squeue_status_string]
            + [self.mock_scontrol_status_string]
        )
        # make _check_sacct_is_active_on_server return False to enable scontrol command
        self.slurm_job_status_cache._check_if_sacct_is_disabled_on_server = lambda: True
        self.assertEqual(self.slurm_job_status_cache._check_if_sacct_is_disabled_on_server(), True)
        # Run assertion test
        self.slurm_job_status_cache._ask_for_job_status()
        self.assertEqual(mock_check_output.call_count, 3)

    @mock.patch("subprocess.check_output")
    def test_ask_for_job_with_sqeue_sacct_raises_assertion_for_incorrect_formatted_output(self, mock_check_output):
        """Test that when a slurm system returns an unexpected return string, an assertion is raised"""
        mock_check_output.side_effect = [self.mock_incorrect_sqeue_sacct_status_string]
        # Run assertion test
        with self.assertRaises(AssertionError):
            self.slurm_job_status_cache._ask_for_job_status()



import b2luigi
import os


class MyTask(b2luigi.DispatchableTask):
    def output(self):
        yield self.add_to_output("some_file.txt")
        yield self.add_to_output("some_other_file.txt")

    def process(self):
        print("Hello!")
        with open(self.get_output_file_name("some_file.txt"), "w") as f:
            f.write("Done")

        print("Bye!")
        import sys

        sys.stdout.flush()
        os.kill(os.getpid(), 11)

        with open(self.get_output_file_name("some_other_file.txt"), "w") as f:
            f.write("Done")


if __name__ == "__main__":
    b2luigi.set_setting("result_dir", "results")
    b2luigi.process(MyTask())



import b2luigi


class ROOTLocalTarget(b2luigi.LocalTarget):
    class ROOTLocalTarget:
        """
        A custom target class that extends :class:`b2luigi.LocalTarget` to handle ROOT files.

        This class overrides the ``exists`` method to not only check for the existence
        of the file but also verify that the ROOT file contains at least one key.
        """

    def exists(self):
        """
        Checks if the target file exists and contains at least one key.

        This method first verifies the existence of the file using the parent class's
        ``exists`` method. If the file exists, it opens the file using ROOT's TFile,
        retrieves the list of keys, and checks if the list contains any entries.

        Returns:
            bool: True if the file exists and contains at least one key, False otherwise.
        """
        if not super().exists():
            return False

        path = self.path

        import ROOT

        tfile = ROOT.TFile.Open(path)
        len_list_of_keys = len(tfile.GetListOfKeys())
        tfile.Close()
        return len_list_of_keys > 0



import subprocess
import sys

import b2luigi

from unittest import TestCase
import tempfile
import os
import shutil


class B2LuigiTestCase(TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        self.cwd = os.getcwd()
        os.chdir(self.test_dir)

        b2luigi.set_setting("result_dir", os.path.join(self.test_dir, "results"))

    def tearDown(self):
        os.chdir(self.cwd)
        shutil.rmtree(self.test_dir)

    def call_file(self, file_name, cwd=None, cli_args=None, **kwargs):
        if not cwd:
            cwd = self.test_dir
        if not cli_args:
            cli_args = []
        test_file_name = os.path.join(os.path.dirname(__file__), file_name)
        tmp_file_name = os.path.join(self.test_dir, os.path.basename(file_name))
        shutil.copy(test_file_name, tmp_file_name)
        return subprocess.check_output([sys.executable, tmp_file_name, *cli_args], cwd=cwd, **kwargs)



from ..helpers import B2LuigiTestCase


class TestDryRun(B2LuigiTestCase):
    def test_dry_run_unfinished(self):
        output = self.call_file(
            "cli/process_dry_run_unfinished.py",
            cli_args=["--dry-run"],
        )
        self.assertIn(b"This is a dummy output.", output)

    def test_dry_run_finished(self):
        output = self.call_file(
            "cli/process_dry_run_finished.py",
            cli_args=["--dry-run"],
        )
        self.assertIn(b"All tasks are finished!", output)



import abc
from cachetools import TTLCache


class BatchJobStatusCache(abc.ABC, TTLCache):
    """
    Abstract base class for job status caches.
    Useful if the batch system provides the status of all jobs
    as a list, which might be faster than asking for each job
    separately.

    Override the function ``_ask_for_job_status``, which should
    set the job status for the specific job if
    specified or for all accessible jobs (e.g. for all of this user).
    Having too much information (e.g. information on jobs
    which are not started by this b2luigi instance) does not matter.
    """

    def __init__(self):
        super().__init__(maxsize=1000, ttl=20)

    @abc.abstractmethod
    def _ask_for_job_status(self, job_id=None):
        """
        Abstract method to query the status of a job.

        This method must be implemented by subclasses to define how to
        retrieve the status of a job based on its unique identifier.

        Args:
            job_id (str, optional): The unique identifier of the job. Defaults to ``None``.
        """
        pass

    def __missing__(self, job_id):
        """
        Handle missing keys in the cache.

        This method is called when a key (``job_id``) is not found in the cache.
        It attempts to retrieve the job status by first querying for all jobs
        and then specifically for the missing job. If the job is found during
        these queries, it is added to the cache and returned.

        Args:
            job_id (str): The identifier of the job to retrieve.

        Returns:
            The job status associated with the given ``job_id`` if found.

        Raises:
            KeyError: If the ``job_id`` cannot be found after querying.
        """
        # First, ask for all jobs
        self._ask_for_job_status(job_id=None)
        if job_id in self:
            return self[job_id]

        # Then, ask specifically for this job
        self._ask_for_job_status(job_id=job_id)
        if job_id in self:
            return self[job_id]

        raise KeyError
